"""Instruments which receive probe events."""

import datetime
import sys
import six

try:
    from statsd import statsd
except ImportError:
    statsd = None

import diagnose
from diagnose import probes


omitted = object()


class Instrument(object):
    """An instrument which receives FunctionProbe events.

    Attributes:
        * name: a name for the instrument; may be used in output, such as
                when constructing statsd metric names.
        * value: a Python expression to be evaluated; its result is the
                 "process variable" to be used as the instrument sees fit.
        * event: a string declaring when to fire the instrument, one of:
            * call: evaluate the value just before calling the wrapped
                    function, in a context which contains the local variables:
                * start: float time
                * now: datetime.datetime.utcnow()
                * args/kwargs: inputs to the target function; these are
                               also included in locals() by their argnames.
                * frame: sys._getframe() of the patch wrapper
             * return: the default; evaluate the value just after the wrapped
                       function returns, in a context with the additional locals:
                * result: the return value of the target function
                * end/elapsed: float times
             * end: evaluate the value in the context of the wrapped function
                    just before it returns.
        * expires: a datetime, after which point the instrument will not fire,
                   or None to mean no expiration
        * custom: a dict of any additional data for subclasses. May include
                  other information for filtering events, set points for
                  closed-loop controllers, or other information specific
                  to the kind of instrument.
        * mgr: an InstrumentManager instance. If None, defaults to the global
               diagnose.manager
    """

    error_expiration = datetime.datetime(1970, 1, 1)

    def __init__(
        self, name, value, event="return", expires=None, custom=None, mgr=None, **kwargs
    ):
        if mgr is None:
            mgr = diagnose.manager
        self.mgr = mgr
        self.name = name
        self.value = value
        self.event = event
        self.expires = expires
        self.custom = custom or {}

    def __str__(self):
        return "%s(name=%r, value=%r, event=%r, expires=%r, custom=%r)" % (
            self.__class__.__name__,
            self.name,
            self.value,
            self.event,
            self.expires,
            self.custom,
        )

    __repr__ = __str__

    def evaluate(self, value, _globals, _locals):
        # Skip eval() if a local variable name
        v = _locals.get(value, omitted)
        if v is omitted:
            v = eval(value, _globals, _locals)
        return v

    def merge_tags(self, _globals, _locals):
        tags = self.mgr.get_tags()
        tag_expr = self.custom.get("tags", None)
        if tag_expr:
            t = self.evaluate(tag_expr, _globals, _locals)
            if isinstance(t, dict):
                t = ["%s:%s" % pair for pair in six.iteritems(t)]
            if not isinstance(t, list):
                raise TypeError("Cannot send non-list of tags: %s" % (t,))
            tags = tags + t
        return tags

    def fire(self, _globals, _locals):
        raise NotImplementedError()

    def check_call(self, probe, *args, **kwargs):
        """Return True if this instrument should be applied, False otherwise.

        By default, this always returns True. Override this in a subclass
        to check the supplied function args/kwargs, or other state,
        such as self.custom, environment variables, or threadlocals.
        """
        return self.mgr.check_call(probe, self, *args, **kwargs)

    def handle_error(self, probe):
        if self.error_expiration:
            # Set self.expires to long ago, which keeps it from firing until:
            # a) someone edits the probe, or
            # b) processes restart, which could be new code that fixes things.
            # Even if it doesn't, we only get ~1 error per process,
            # not 1 per call to the target function.
            self.expires = self.error_expiration
        self.mgr.handle_error(probe, self)

    def __call__(self, f):
        """Use self as a decorator, attaching a probe to the wrapped function."""
        classname = sys._getframe(1).f_code.co_name
        if classname == "<module>":
            target = "%s.%s" % (f.__module__, getattr(f, "__name__", None) or getattr(f, "func_name"))
        else:
            target = "%s.%s.%s" % (f.__module__, classname, getattr(f, "__name__", None) or getattr(f, "func_name"))

        probe = probes.attach_to(target)
        # If we prefix the spec_id with self.mgr.short_id, then that
        # manager would immediately remove this instrument because
        # it's not in self.mgr.specs!
        # Use a hardcoded prefix instead so no manager drops it.
        probe.instruments["hardcode:%s" % (hash(target),)] = self

        return f


class LogInstrument(Instrument):
    """An instrument that prints a log message."""

    MAX_CHARS = 2000
    out = sys.stdout

    def fire(self, _globals, _locals):
        v = self.evaluate(self.value, _globals, _locals)
        if v is None:
            return

        v = str(v)
        if len(v) > self.MAX_CHARS:
            v = v[: self.MAX_CHARS - 3] + "..."

        tags = self.merge_tags(_globals, _locals)

        t = str(tags)
        if len(t) > self.MAX_CHARS:
            t = t[: self.MAX_CHARS - 3] + "..."

        self.emit(self.name, v, t)

    def emit(self, name, value, tags):
        self.out.write("Probe (%s)[tags=%s] = %s\n" % (name, tags, value))


class StatsdInstrumentBase(Instrument):
    """An instrument that sends a value to statsd."""

    MAX_CHARS = 2000

    def fire(self, _globals, _locals):
        v = self.evaluate(self.value, _globals, _locals)
        if v is None:
            return

        if not isinstance(v, (int, float, long)):
            v = str(v)
            if len(v) > self.MAX_CHARS:
                v = v[: self.MAX_CHARS] + "..."
            raise TypeError("Cannot send non-numeric metric: %s" % (v,))

        self.emit(self.name, v, self.merge_tags(_globals, _locals))

    def emit(self, name, value, tags):
        raise NotImplementedError()


class HistogramInstrument(StatsdInstrumentBase):
    def emit(self, name, value, tags):
        statsd.histogram(name, value, tags=tags)


class IncrementInstrument(StatsdInstrumentBase):
    def emit(self, name, value, tags):
        statsd.increment(name, value, tags=tags)


class ProbeTestInstrument(Instrument):
    """An instrument that stores values in self.results."""

    def __init__(self, *args, **kwargs):
        Instrument.__init__(self, *args, **kwargs)
        self.log = []

    @property
    def results(self):
        return [result for tags, result in self.log]

    def fire(self, _globals, _locals):
        v = self.evaluate(self.value, _globals, _locals)
        tags = self.merge_tags(_globals, _locals)
        self.log.append((tags, v))



"""Wrappers to monitor function execution."""

from collections import namedtuple
import datetime
import functools
import gc
import linecache
import six
import sys
import time
import traceback
import types
import weakref

import hunter
import mock


omitted = object()

# A registry of FunctionProbe instances.
# Since probes patch functions, it's imperative that the same function
# not be patched twice, even in test suites or other scenarios where
# two services might be applying probes at the same time. This global
# registry enforces that. Callers should add instruments to probes instead.
# The FunctionProbe class itself uses this to disallow attaching two
# FunctionProbes to the same target.
active_probes = {}


def attach_to(target):
    """Return the probe attached to the given target, or a new one if needed."""
    probe = active_probes.get(target, None)
    if probe is None:
        probe = FunctionProbe(target)
        active_probes[target] = probe
    return probe


def start_all():
    for probe in six.itervalues(active_probes):
        probe.start()


class FunctionProbe(object):
    """A wrapper for a function, to monitor its execution.

    target: a dotted Python path to the function to wrap.
    instruments: a dict of ("spec id": Instrument instance) pairs.

    When started, a FunctionProbe "monkey-patches" its target, replacing
    it with a wrapper function. That wrapper calls the original target
    function, calling each instrument at the event it specifies: call, end,
    or return. If the event is "end", then the probe registers a trace function
    (via `sys.settrace`) and then calls those instruments just before the
    original target function returns; the instrument values are evaluated
    in the globals and locals of that frame. "Call" and "return" instruments
    are evaluated before or after the function returns; they are much less
    invasive since they do not call settrace, but they are limited to reading
    the function's inputs, outputs, and timing.
    """

    def __init__(self, target, instruments=None):
        if target in active_probes:
            raise RuntimeError(
                "Cannot apply two probes to the same target. "
                "Try calling attach_to(target) instead of FunctionProbe(target)."
            )
        self.target = target
        if instruments is None:
            instruments = {}
        self.instruments = instruments
        self.patches = {}
        active_probes[target] = self

    def __str__(self):
        return "%s(target=%r, instruments=%r)" % (
            self.__class__.__name__,
            self.target,
            self.instruments,
        )

    __repr__ = __str__

    def make_patches(self):
        """Set self.patches to a list of mock._patch objects which wrap our target function."""
        primary_patch = mock.patch(self.target)

        # Replace the target with a wrapper.
        original, local = primary_patch.get_original()
        if isinstance(original, types.FunctionType):
            base = original
        elif isinstance(original, (staticmethod, classmethod)):
            base = original.__func__
        elif isinstance(original, property):
            base = original.fget
        else:
            raise TypeError(
                "Cannot probe: %s is not a function." % (repr(self.target),)
            )
        varnames = self.maybe_unwrap(base).__code__.co_varnames

        @functools.wraps(base)
        def probe_wrapper(*args, **kwargs):
            now = datetime.datetime.utcnow()

            hotspots = HotspotsFinder()
            instruments_by_event = {"call": [], "return": [], "end": []}
            for I in six.itervalues(self.instruments):
                if I.expires and now > I.expires:
                    continue
                if I.check_call(self, *args, **kwargs):
                    instruments_by_event[I.event].append(I)
                    if I.event in ("call", "return"):
                        if "hotspots" in I.value or "hotspots" in (
                            I.custom.get("tags") or ""
                        ):
                            hotspots.enabled = True

            target_obj, target_func_name = self.target.rsplit(".", 1)
            is_unwrapped = base.__code__.co_name == target_func_name
            if instruments_by_event["end"]:
                # We have instruments that require evaluation in the local
                # context of the function. Call sys.settrace() to gain access.
                predicate = hunter.When(
                    hunter.Query(
                        # Only trace returns (this will include exceptions!)...
                        kind="return",
                        # ...and only in the given function...
                        function=target_func_name,
                        # ...(no deeper).
                        depth=0,
                    )
                    if is_unwrapped
                    else hunter.Query(
                        # Only trace returns (this will include exceptions!)...
                        kind="return",
                        # ...and only in the given function...
                        function=target_func_name,
                        # ...but we don't know how many times it's been wrapped.
                        # Use the module instead as an approximate match.
                        # This may catch other functions with the same name
                        # in the same module, but not much we can do about
                        # that without a custom Cython Query.
                        module_in=target_obj,
                    ),
                    TraceHandler(self, instruments_by_event["end"]),
                )
                tracer = hunter.Tracer(
                    # There's no need to call threading.settrace() because
                    # a) we're targeting a function we're about to call
                    #    in the same thread,
                    # b) we're going to undo it immediately after, and
                    # c) it would collide with other threads if they did
                    #    the same concurrently.
                    threading_support=False
                ).trace(predicate)
            elif hotspots.enabled:
                # We have instruments that require timing internal lines.
                # Call sys.settrace() to gain access.
                predicate = hunter.When(
                    hunter.Query(
                        # Only trace lines...
                        kind="line",
                        # ...and only in the given function...
                        function=target_func_name,
                        # ...(no deeper).
                        depth=1,
                    )
                    if is_unwrapped
                    else hunter.Query(
                        # Only trace lines...
                        kind="line",
                        # ...and only in the given function...
                        function=target_func_name,
                        # ...but we don't know how many times it's been wrapped.
                        # Use the module instead as an approximate match.
                        # This may catch other functions with the same name
                        # in the same module, but not much we can do about
                        # that without a custom Cython Query.
                        module_in=target_obj,
                    ),
                    hotspots,
                )
                tracer = hunter.Tracer(
                    # There's no need to call threading.settrace() because
                    # a) we're targeting a function we're about to call
                    #    in the same thread,
                    # b) we're going to undo it immediately after, and
                    # c) it would collide with other threads if they did
                    #    the same concurrently.
                    threading_support=False
                ).trace(predicate)
            else:
                tracer = None

            try:
                if instruments_by_event["call"] or instruments_by_event["return"]:
                    start = time.time()
                    _locals = {
                        "start": start,
                        "now": now,
                        "args": args,
                        "kwargs": kwargs,
                        "frame": sys._getframe(),
                    }
                    # Add positional args to locals by name.
                    for i, argname in enumerate(varnames[: len(args)]):
                        if argname not in ("args", "kwargs"):
                            _locals[argname] = args[i]
                    # Add kwargs to locals
                    _locals.update(kwargs)

                for instrument in instruments_by_event["call"]:
                    try:
                        instrument.fire(instrument.mgr.global_namespace, _locals)
                    except:
                        try:
                            instrument.handle_error(self)
                        except:
                            traceback.print_exc()

                # Execute the base function and obtain its result.
                try:
                    result = base(*args, **kwargs)
                except:
                    result = sys.exc_info()[1]
                    raise
                finally:
                    if hotspots.enabled:
                        hotspots.finish()
                        _locals["hotspots"] = hotspots

                    if instruments_by_event["return"]:
                        end = time.time()
                        elapsed = end - start
                        _locals.update(
                            {"result": result, "end": end, "elapsed": elapsed}
                        )

                    for instrument in instruments_by_event["return"]:
                        try:
                            instrument.fire(instrument.mgr.global_namespace, _locals)
                        except:
                            try:
                                instrument.handle_error(self)
                            except:
                                traceback.print_exc()

                return result
            finally:
                if tracer is not None:
                    tracer.stop()

        if isinstance(original, property):
            # We can't patch original.fget directly because it's read-only,
            # so we replace the whole property with a new one, passing our
            # probe_wrapper as its fget.
            # At this time, we only patch fget. If there's enough demand,
            # we could do all three in the future, but then that would take
            # three probe_wrapper functions, and what the instruments do
            # with three instead of one could be very confusing.
            primary_patch.new = property(
                probe_wrapper, original.fset, original.fdel, original.__doc__
            )
        else:
            if isinstance(original, staticmethod):
                probe_wrapper = staticmethod(probe_wrapper)
            elif isinstance(original, classmethod):
                probe_wrapper = classmethod(probe_wrapper)
            primary_patch.new = probe_wrapper

        patches = {0: primary_patch}

        # Add patches for any other modules/classes which have
        # the target as an attribute, or "registry" dicts which have
        # the target as a value.
        _resolved_target = primary_patch.getter()
        for ref in gc.get_referrers(original):
            if not isinstance(ref, dict):
                continue

            names = [k for k, v in ref.items() if v is original]
            for parent in gc.get_referrers(ref):
                if parent is _resolved_target or parent is primary_patch:
                    continue

                if getattr(parent, "__dict__", None) is ref:
                    # An attribute of a module or class or instance.
                    for name in names:
                        patch_id = len(patches)
                        patch = WeakMethodPatch(
                            self.make_getter(patch_id, parent), name, probe_wrapper
                        )
                        patches[patch_id] = patch
                else:
                    for gpa in gc.get_referrers(parent):
                        if getattr(gpa, "__dict__", None) is parent:
                            # A member of a dict which is
                            # an attribute of a module or class or instance.
                            for name in names:
                                patch_id = len(patches)
                                patch = DictPatch(ref, name, probe_wrapper)
                                patches[patch_id] = patch
                            break

        self.patches = patches

    @staticmethod
    def maybe_unwrap(func):
        """Return the given function, without its probe_wrapper if it has one."""
        if getattr(getattr(func, "__code__", None), "co_name", "") == "probe_wrapper":
            return func.__closure__[0].cell_contents
        else:
            try:
                # If the given func is a func returned from @functools.wraps(orig),
                # then the only cell in its closure will be the `orig` function.
                # There may be funcs with a single cell that is a function
                # that is not the wrapped function, in which case we will
                # return the wrong signature here, but use of wraps is so
                # much more common and SO useful to unwrap that we risk it.
                if len(func.__closure__) == 1:
                    f = func.__closure__[0].cell_contents
                    if hasattr(f, "__code__"):
                        return f
            except (AttributeError, IndexError, TypeError):
                pass

        return func

    def make_getter(self, patch_id, parent):
        def callback(ref):
            p = self.patches.get(patch_id, None)
            if p:
                try:
                    p.stop()
                except RuntimeError:
                    # Already stopped. Ignore.
                    pass
                self.patches.pop(patch_id, None)

        try:
            getter = weakref.ref(parent, callback)
        except TypeError:

            def getter():
                return parent

        return getter

    def start(self):
        """Apply self.patches. Safe to call after already started."""
        if not self.patches:
            self.make_patches()
        for p in six.itervalues(self.patches):
            if not hasattr(p, "is_local"):
                p.start()

    def stop(self):
        for p in six.itervalues(self.patches):
            try:
                p.stop()
            except RuntimeError:
                # Already stopped. Ignore.
                pass


class TraceHandler(object):
    """A sys.settrace arg, which calls instruments in the context of the frame."""

    def __init__(self, probe, instruments):
        self.probe = probe
        self.instruments = instruments

    def __call__(self, event):
        _globals = event.globals
        _locals = {"__event__": event}
        _locals.update(event.locals)
        for instrument in self.instruments:
            try:
                _g = _globals.copy()
                _g.update(instrument.mgr.global_namespace)
                instrument.fire(_g, _locals)
            except:
                try:
                    instrument.handle_error(self.probe)
                except:
                    traceback.print_exc()


CallTime = namedtuple("CallTime", ["time", "lineno", "source"])


class HotspotsFinder(object):
    """A sys.settrace arg, which records line timings."""

    def __init__(self):
        self.enabled = False
        self._last_time = None
        self._last_line = None
        self.calls = {}
        self.filename = None

    def __call__(self, event=None):
        if self._last_time is not None:
            elapsed = time.time() - self._last_time
            ll = self._last_line
            call = self.calls.get(ll, None)
            if call is None:
                # count, max, sum
                self.calls[ll] = [1, elapsed, elapsed]
            else:
                call[0] += 1
                if elapsed > call[1]:
                    call[1] = elapsed
                call[2] += elapsed

        if event is not None:
            self._last_line = event.lineno
            if self.filename is None:
                self.filename = event.filename
            # Don't include this method's time in the next line time
            self._last_time = time.time()

    def finish(self):
        # Fake the last line time
        self.__call__(event=None)

        if self.calls:
            worst = max(((call[2], lineno) for lineno, call in six.iteritems(self.calls)))
            self.worst = CallTime(*worst, source=self.source(worst[1]))

            slowest = max(
                ((call[1], lineno) for lineno, call in six.iteritems(self.calls))
            )
            self.slowest = CallTime(*slowest, source=self.source(slowest[1]))
        else:
            self.worst = self.slowest = CallTime(None, None, None)

    def source(self, lineno):
        return linecache.getline(self.filename, lineno)


# ----------------------------- Weak patch ----------------------------- #


class WeakMethodPatch(object):
    """A Patch for an attribute a Python object.

    On start/__enter__, calls self.getter() which should return an object,
    then replaces the given attribute of that object with the new value.
    On stop/__exit__, replaces the same attribute with the previous value.

    Used by FunctionPatch to replace references to functions which appear in
    modules, classes, or other objects. Weak references are used internally
    so that, if the object is removed from that module etc (has no more strong
    references), then the patch is automatically abandoned.
    """

    def __init__(self, getter, attribute, new):
        self.getter = getter
        self.attribute = attribute
        self.new = new

    def get_original(self):
        target = self.getter()
        name = self.attribute

        original = omitted
        local = False

        try:
            original = target.__dict__[name]
        except (AttributeError, KeyError):
            original = getattr(target, name, omitted)
        else:
            local = True

        if original is omitted:
            raise AttributeError("%s does not have the attribute %r" % (target, name))
        return original, local

    def __enter__(self):
        """Perform the patch."""
        original, local = self.get_original()
        self.temp_original = weakref.ref(original)
        self.is_local = local
        setattr(self.getter(), self.attribute, self.new)
        return self.new

    def __exit__(self, *exc_info):
        """Undo the patch."""
        if not hasattr(self, "is_local"):
            raise RuntimeError("stop called on unstarted patcher")

        target = self.getter()
        if target is None:
            return

        original = self.temp_original()
        if original is None:
            return

        if getattr(target, self.attribute, None) is self.new:
            if self.is_local:
                setattr(target, self.attribute, original)
            else:
                delattr(target, self.attribute)
                if not hasattr(target, self.attribute):
                    # needed for proxy objects like django settings
                    setattr(target, self.attribute, original)

        del self.is_local

    def start(self):
        """Activate a patch, returning any created mock."""
        result = self.__enter__()
        return result

    def stop(self):
        """Stop an active patch."""
        return self.__exit__()


class DictPatch(object):
    """A Patch for a member of a Python dictionary.

    On start/__enter__, replaces the member of the given dictionary
    identified by the given key with a new object. On stop/__exit__,
    replaces the same key with the previous object.

    Used by FunctionPatch to replace references to functions which appear
    in any dictionary, such as a function registry.
    """

    def __init__(self, dictionary, key, new):
        self.dictionary = dictionary
        self.key = key
        self.new = new

    def get_original(self):
        return self.dictionary[self.key], True

    def __enter__(self):
        """Perform the patch."""
        original, local = self.get_original()
        self.temp_original = original
        self.is_local = local
        self.dictionary[self.key] = self.new
        return self.new

    def __exit__(self, *exc_info):
        """Undo the patch."""
        if not hasattr(self, "is_local"):
            raise RuntimeError("stop called on unstarted patcher")

        self.dictionary[self.key] = self.temp_original

        del self.is_local

    def start(self):
        """Activate a patch, returning any created mock."""
        result = self.__enter__()
        return result

    def stop(self):
        """Stop an active patch."""
        return self.__exit__()



"""Simulate concurrent schedules by pausing execution at a given function."""
import functools
import threading
import time
import inspect

omitted = object()


class Breakpoint:
    """A tool to synchronize execution or simulate errors during tests.

    Use a Breakpoint in tests to simulate deterministic interleavings of
    operations which would otherwise be left to concurrency accidents
    or complicated calls to sleep. You can also simulate internal errors
    with the `throw` argument. All of this can be done entirely in tests,
    without infecting production code with a bunch of scaffolding.

    A Breakpoint acts like a semaphore, allowing a test and the system
    (the thread being tested) to synchronize their execution. The test calls
    bp.wait(), which blocks; when the system hits the breakpoint, they switch:
    the system blocks and the test unblocks. While the system is blocked,
    the test is free to inspect or alter state without fear. For example:

        with Breakpoint(obj, funcname) as bp:
            bp.start_thread(foo)  # start running foo in a thread
            bp.wait()             # wait for that thread to hit the function

            assert foo.is_running
            foo.set_interrupt()

    If the system somehow hits the breakpoint before the test calls wait(),
    then wait() simply returns immediately. If the system doesn't hit
    the breakpoint within the timeout, wait() throws a RuntimeError.

    The test may unblock the system by explicitly calling bp.release(),
    or just waiting for the "with" block to exit (if neither happens
    within the timeout, the system thread throws RuntimeError).

    Breakpoints may be conditional based on the arguments being passed.
    Set bp.condition to a callable which takes the same arguments as the
    breakpoint function; if it returns True, the call blocks and control
    is passed back to the test thread; if False, the call proceeds normally.
    Alternately, the condition may be an int or list of ints, in which case
    it will break on those numbered calls.

    If the `throw` argument is provided, it should be an exception instance,
    which is raised whenever the condition is met instead of blocking.
    Use this to simulate errors at specific points rather than simulating
    deterministic interleaving of operations.

    When a breakpoint is hit, the breakpoints "stackframe" attribute is
    set to the current frame. Using this, you can inspect the call
    stack or function arguments.
    """

    def __init__(
        self,
        obj,
        funcname,
        condition=None,
        timeout=10.0,
        break_on_return=False,
        throw=None,
        check_interval=0.1,
    ):
        """
        obj:
            Class or object to be patched
        funcname:
            str: Name of method/function on that object
        condition:
            None to always break when function is called, or callable that
            takes same args as patched function and returns True to break,
            False to keep going. Alternately, it may be an int or list of
            ints, in which case it will break on those numbered calls.
        break_on_return:
            False (default) to break when the function is entered,
            True to break when the function exits.
        """
        self.blocked = 0
        self.condition = condition
        self.timeout = timeout
        self.break_on_return = break_on_return
        self.throw = throw
        self.check_interval = check_interval
        self._started_threads = []
        self.stackframe = None

        self.obj = obj
        self.funcname = funcname
        self.calls = []
        f = getattr(obj, funcname)

        @functools.wraps(f)
        def breakpoint_wrapper(*args, **kwargs):
            self.stackframe = inspect.currentframe()
            if not self.break_on_return:
                self._block_if_condition_met(args, kwargs)

            result = f(*args, **kwargs)

            if self.break_on_return:
                self._block_if_condition_met(args, kwargs)

            # Best practice is not to hold onto stackframe longer
            # than it is needed.
            self.stackframe = None
            return result

        self.original = f
        self.wrapper = breakpoint_wrapper

    def _block_if_condition_met(self, args, kwargs):
        if self.condition is None:
            met = True
        elif isinstance(self.condition, int):
            met = len(self.calls) == self.condition
        elif isinstance(self.condition, (set, tuple, list)):
            met = len(self.calls) in self.condition
        elif callable(self.condition):
            met = self.condition(*args, **kwargs)
        else:
            raise TypeError(
                "Breakpoint.condition must be None, an int or list of ints, or a callable."
            )
        self.calls.append(met)

        if met:
            if self.throw is not None:
                raise self.throw

            start = time.time()
            timeout = self.timeout
            self.blocked += 1
            while self.blocked:
                if timeout is not None and time.time() - start > timeout:
                    raise RuntimeError(
                        "Breakpoint on %s.%s timed out after %s seconds."
                        % (self.obj, self.funcname, timeout)
                    )
                time.sleep(self.check_interval)

    def __enter__(self):
        self.calls = []
        self._started_threads = []
        setattr(self.obj, self.funcname, self.wrapper)
        self.blocked = 0
        return self

    def __exit__(self, type, value, traceback):
        setattr(self.obj, self.funcname, self.original)
        self.blocked = 0
        for t in self._started_threads:
            t.join()
        self._started_threads = []

        if type is not None:
            # There was already an error, don't suppress it.
            return False

        if isinstance(self.condition, int):
            if len(self.calls) <= self.condition or not self.calls[self.condition]:
                raise AssertionError(
                    "Breakpoint condition on %s was not met for iteration %s."
                    % (self.funcname, self.condition)
                )
        elif isinstance(self.condition, (set, tuple, list)):
            not_called = [
                c for c in self.condition if len(self.calls) <= c or not self.calls[c]
            ]
            if not_called:
                raise AssertionError(
                    "Breakpoint condition on %s was not met for iterations %s."
                    % (self.funcname, not_called)
                )
        else:
            if not any(self.calls):
                raise AssertionError(
                    "Breakpoint condition on %s was not met." % (self.funcname,)
                )

    def start_thread(self, target, **kwargs):
        """Execute the given target (a callable) in another thread.

        Tests may call this to run the system in the background.

        Threads started here will be joined on breakpoint context exit
        (but only *after* unblocking).
        """
        t = threading.Thread(target=target, **kwargs)
        self._started_threads.append(t)
        t.start()

    def join(self):
        """
        Waits for all the threads to complete.
        """
        for thread in self._started_threads:
            thread.join()

    def wait(self, timeout=omitted, hits=1):
        """Block until the breakpoint is hit by other thread(s).

        Tests should call this to wait for the system to advance to the
        breakpoint function. Once it does, this method returns and the
        test may proceed while the system is blocked. Call release()
        to unblock the system again.

        If `hits` is greater than 1, this function waits until the breakpoint
        has been hit the given number of times before returning.
        """
        if timeout is omitted:
            timeout = self.timeout

        start = time.time()
        while self.blocked < hits:
            if timeout is not None and time.time() - start > timeout:
                raise RuntimeError(
                    "Breakpoint on %s not hit after %s seconds."
                    % (self.funcname, timeout)
                )
            time.sleep(self.check_interval)

    def wait_until(self, condition, timeout=omitted):
        """Block until the condition is True, or error if the timeout is reached.

        Call this to wait for the system to advance to a certain state;
        once the given `condition` callable returns True, this function
        returns. If the timeout (or self.timeout if omitted) is reached,
        a RuntimeError is raised.
        """
        if timeout is omitted:
            timeout = self.timeout

        start = time.time()
        while not condition():
            if timeout is not None and time.time() - start > timeout:
                raise RuntimeError(
                    "Condition for %s not met after %s seconds."
                    % (self.funcname, timeout)
                )
            time.sleep(self.check_interval)

    def release(self):
        """Allow the system to proceed (until the breakpoint is hit again).

        When the system hits this breakpoint, it blocks while the test
        proceeds. The test should call this method to unblock the system.
        """
        self.blocked = 0



"""Diagnose, a library for instrumenting code at runtime."""

from contextlib import contextmanager

from . import managers
from . import instruments
from . import probes

# A global since it should be one per process.
# You _may_ make another, but most people will just want the one.
# You should probably `import diagnose` and then refer to `diagnose.manager`
# rather than `from diagnose import manager` in case some framework
# decides to replace diagnose.manager with an instance of a subclass.
manager = managers.InstrumentManager()

__all__ = ("probes", "instruments", "manager", "managers", "sensor")


@contextmanager
def sensor(target, value="result", name="test_instrument", event="return", mgr=None):
    """Attach a probe to the given target and yield a ProbeTestInstrument."""
    probe = probes.attach_to(target)
    probe.instruments[name] = i = instruments.ProbeTestInstrument(
        name, value, event, expires=None, mgr=mgr
    )
    probe.start()

    yield i

    probe.instruments.pop(name)
    if not probe.instruments:
        p = probes.active_probes.pop(target, None)
        if p is not None:
            p.stop()



"""Managers for probe instruments."""

import datetime
import math
import six
import sys
import threading
import time
import traceback

from . import instruments
from . import probes


def mag(obj, base=10):
    """Return the magnitude of the object (or its length), or -1 if size is 0."""
    try:
        size = len(obj)
    except TypeError:
        size = obj

    try:
        m = int(math.log10(size) if base == 10 else math.log(size, base))
    except:
        m = -1

    return m


# --------------------------- Instrument Manager --------------------------- #


class InstrumentManager(object):
    """A Manager which applies instruments to probes to targets.

    specs: a dict of {id: spec} dicts, each of which defines an instrument. Fields:
        * target: the dotted path to a function to wrap in a FunctionProbe
        * lastmodified: the datetime when the spec was last changed
        * lifespan: an int; the number of minutes from lastmodified before
                    the instrument should expire.
        * instrument: a dict of Instrument params (type, name, value, event, custom).
                      The "expires" param is calculated from lastmodified + lifespan.
        * applied: a dict of {process_id: info} pairs, where "info" is a dict with:
            * lm: the "lastmodified" datetime of the instrument when
                 it was applied to the given process. Monitors can use
                 this to know when processes have the latest spec or not.
            * err: a string with any error raised in the given process, or None
    """

    instrument_classes = {
        "log": instruments.LogInstrument,
        "hist": instruments.HistogramInstrument,
        "incr": instruments.IncrementInstrument,
    }
    lock = threading.Lock()

    # Adjust here any modules or other globals you want instruments to access.
    # Instrument.value could of course __import__("whatever") but that is slow.
    global_namespace = {"datetime": datetime, "math": math, "time": time, "mag": mag}

    def __init__(self, process_id=None):
        self.specs = {}
        self.process_id = process_id
        self.apply_thread = None
        self.period = 60
        self.short_id = hex(id(self))[2:]

    def apply(self):
        """Add/remove instruments to match our spec."""
        with self.lock:
            self._apply()

    def _apply(self):
        seen_instruments = {}
        for spec_id, doc in six.iteritems(self.specs):
            full_id = "%s:%s" % (self.short_id, spec_id)
            target = doc["target"]
            seen_instruments[full_id] = target
            probe, I, cls, expires = None, None, None, None
            try:
                probe = probes.attach_to(target)

                lifespan = doc["lifespan"]
                if lifespan is None:
                    expires = None
                else:
                    expires = doc["lastmodified"] + datetime.timedelta(minutes=lifespan)

                cls = self.instrument_classes[doc["instrument"]["type"]]

                # Add or modify instruments
                I = probe.instruments.get(full_id, None)
                modified = False
                if I is None or I.__class__ != cls:
                    probe.instruments[full_id] = cls(
                        mgr=self, expires=expires, **doc["instrument"]
                    )
                    modified = True
                else:
                    for key in ("name", "value", "event", "custom"):
                        if getattr(I, key) != doc["instrument"][key]:
                            setattr(I, key, doc["instrument"][key])
                            modified = True
                    if I.expires != expires:
                        I.expires = expires
                        modified = True
                if modified:
                    self.mark(spec_id, doc)
            except:
                if I is None:
                    if cls is None:
                        cls = self.instrument_classes[doc["instrument"]["type"]]
                    I = cls(mgr=self, expires=expires, **doc["instrument"])
                self.handle_error(probe, I)
                self.mark(spec_id, doc, exception=True)

        # Remove defunct instruments, probes
        for target, probe in probes.active_probes.items():
            for full_id, instrument in probe.instruments.items():
                # Remove any instrument from a probe if this manager
                # thinks it doesn't apply to the same target anymore.
                if full_id.startswith(self.short_id):
                    if seen_instruments.get(full_id, None) != probe.target:
                        probe.instruments.pop(full_id, None)

            if not probe.instruments:
                p = probes.active_probes.pop(target, None)
                if p is not None:
                    p.stop()

        probes.start_all()

    def apply_in_background(self, period=60):
        self.period = period
        if self.apply_thread is None:
            self.apply_thread = t = threading.Thread(target=self._cycle)
            t.setName("diagnose.manager.apply_in_background")
            t.daemon = True
            t.start()

    def _cycle(self):
        while True:
            time.sleep(self.period)
            try:
                self.apply()
            except:
                self.handle_error()

    def check_call(self, probe, instrument, *args, **kwargs):
        """Return True if the given instrument should be called, False otherwise.

        By default, this always returns True. Override this in a subclass
        to check the supplied function args/kwargs, or other state,
        such as instrument.custom, environment variables, or threadlocals.
        """
        return True

    def get_tags(self):
        return []

    def handle_error(self, probe=None, instrument=None):
        """Handle any error raised by an instrument.

        By default, this prints the error.
        Override it to log or whatever else you prefer.
        """
        traceback.print_exc()

    def mark(self, id, doc, exception=False):
        """Record instrument application success/failure."""
        doc = self.specs.get(id, None)
        if doc is not None:
            error = None
            if exception:
                error = "Error: %s\n%s" % (
                    repr(sys.exc_info()[1]),
                    traceback.format_exc(),
                )

            newval = {"lm": doc["lastmodified"], "err": error}
            if doc["applied"].get(self.process_id, {}) != newval:
                doc["applied"][self.process_id] = newval


class MongoDBInstrumentManager(InstrumentManager):
    """An InstrumentManager which reads and writes specs in MongoDB.

    collection: a pymongo collection in which specs are stored.
    """

    def __init__(self, process_id, collection, id_field="id"):
        self.process_id = process_id
        self.apply_thread = None
        self.period = 60
        self.collection = collection
        self.id_field = id_field
        self.short_id = hex(id(self))[2:]

    @property
    def specs(self):
        return dict((doc[self.id_field], doc) for doc in self.collection.find())

    def mark(self, id, doc, exception=False):
        """Record instrument application success/failure."""
        doc = self.specs.get(id, None)
        if doc is not None:
            error = None
            if exception:
                error = "Error: %s\n%s" % (
                    repr(sys.exc_info()[1]),
                    traceback.format_exc(),
                )

            newval = {"lm": doc["lastmodified"], "err": error}
            if doc["applied"].get(self.process_id, {}) != newval:
                doc["applied"][self.process_id] = newval
                self.collection.update(
                    {self.id_field: id},
                    {
                        "$set": {
                            # Oh, Mongo. You and your dots.
                            "applied.%s"
                            % self.process_id.replace(".", "_"): newval
                        }
                    },
                )
